-- Father's Day - a greeting card
-- Copyright (C) 2023 Duane Robertson

-- main.lua

-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.


local FONT_FILE = 'art/roboto/Roboto-Bold.ttf'
local FSIZE = 100
local IHEIGHT = 720
local IWIDTH = 1280

local android
local canv
local canv2
local canvt
local card = {}
local darken
local font
local fontsm
local tq = {}
local scale = { x=1, y=1 }
local sptext


local function twod(n)
	return math.random(n) + math.random(n)
end


function love.draw()
	card:draw()
end


function love.load()
	math.randomseed(os.time())

	canv = love.graphics.newCanvas(IWIDTH, IHEIGHT, { dpiscale = 1 })
	canv2 = love.graphics.newCanvas(IWIDTH, IHEIGHT, { dpiscale = 1 })
	canvt = love.graphics.newCanvas(IWIDTH, IHEIGHT, { dpiscale = 1 })
	canvt2 = love.graphics.newCanvas(IWIDTH, IHEIGHT, { dpiscale = 1 })
	canvi = love.graphics.newCanvas(IWIDTH, FSIZE * 1.5, { dpiscale = 1 })

	font = love.graphics.newFont(FONT_FILE, FSIZE)
	fontsm = love.graphics.newFont(FONT_FILE, 24)

	darken = love.graphics.newShader [[
		vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 pixel_coords)
		{
			float m = 0.02;
			float m2 = 0.7;
			float m3 = 0.4;

			vec4 curr = Texel(texture, texture_coords);
			float l = (curr.r + curr.g + curr.b) / 3.0;

			if (l < m)
				return vec4(0, 0, 0, 0);
			else if (l < m3)
				return vec4(floor(curr.rgb * 95.0) / 100.0, 1.0);
			else if (l < m2)
				return vec4(floor(curr.rgb * 90.0) / 100.0, 1.0);
			else
				return vec4(floor(curr.rgb * 50.0) / 100.0, 1.0);
		}
	]]

	sptext = love.graphics.newShader [[
		uniform float pos;
		uniform Image txtx;
		uniform vec2 txtsz;

		vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 pixel_coords)
		{
			vec2 vpos = pixel_coords - vec2(0.0, pos);
			vec4 txt = Texel(txtx, vpos / txtsz);

			if (txt.a != 0.0)
				return color;
			else
				return vec4(0, 0, 0, 0);
		}
	]]

	card:load()

	android = (love.system.getOS() == 'Android')

	if android then
		local _, _, desk_x, desk_y = love.window.getSafeArea()
		scale.x = desk_x / IWIDTH
		scale.y = desk_y / IHEIGHT
	end
end


function love.mousepressed(x, y, button)
	card.start = true
    if button == 1 then
		card:launch(x / scale.x, y / scale.y)
	elseif button == 2 then
		-- nop
    end
end


function love.update(dt)
	card:update(dt)
end


function card:ballistic(obj, dt)
	obj.dy = obj.dy + 50 * dt
	obj.x = obj.x + obj.dx * dt
	obj.y = obj.y + obj.dy * dt
end


function card:do_circles(dt)
	local i = 1
	while i <= #self.circles do
		local cir = self.circles[i]
		self:ballistic(cir, dt)

		cir.timer = cir.timer - 1
		if cir.fuse then
			cir.fuse = cir.fuse - 1
		end

		if cir.timer < 1 or cir.x < 0
		or cir.x > IWIDTH or cir.y > IHEIGHT then
			tq[#tq+1] = table.remove(self.circles, i)
		elseif (cir.fuse or 1) < 1 then
			cir.fuse = nil
			self:explode(cir)
			tq[#tq+1] = table.remove(self.circles, i)
		else
			i = i + 1
		end
	end
end


function card:do_rockets(dt)
	local i = 1
	while i <= #self.rockets do
		local roc = self.rockets[i]
		if roc.delay > 0 then
			roc.delay = roc.delay - 1
			i = i + 1
		else
			roc.dx = (roc.tx - roc.x) / (math.abs(roc.ty - roc.y) / 300)
			self:ballistic(roc, dt)

			if roc.y - roc.ty < 0 then
				roc.x, roc.y = roc.tx, roc.ty
				local expl = (math.random(3) == 1)
				self:explode(roc, expl)
				table.remove(self.rockets, i)
			elseif roc.x < 0 or roc.x > IWIDTH or roc.y > IHEIGHT + 500 then
				table.remove(self.rockets, i)
			else
				i = i + 1
			end
		end
	end
end


function card:draw()
	self:fade_canvas()
	self:draw_circles()
	self:mask_text()
	self:draw_text()
	self:draw_canvases()
end


function card:draw_canvases()
	love.graphics.setCanvas()
	if android then
		love.graphics.draw(canv, 0, 0, 0, scale.x, scale.y)
		love.graphics.draw(canvt, 0, 0, 0, scale.x, scale.y)
	else
		love.graphics.draw(canv)
		love.graphics.draw(canvt)
	end
end


function card:draw_circles()
	love.graphics.setCanvas(canv)
	love.graphics.setColor({1, 0, 0})

	for _, roc in ipairs(self.rockets) do
		love.graphics.ellipse('fill', roc.x, roc.y, 2, 2)
	end

	for _, cir in ipairs(self.circles) do
		love.graphics.setColor(cir.color)
		love.graphics.ellipse('fill', cir.x, cir.y, 2, 2)
	end
end


function card:draw_text()
	if not self.hfd then
		-- Make an image of the text.
		love.graphics.setCanvas(canvi)
		love.graphics.setFont(font)
		love.graphics.printf('Happy Fathers\' Day!', 0, 0, IWIDTH, 'center')
		love.graphics.setCanvas()
		self.hfd = canvi:newImageData()
		self.hfd = love.graphics.newImage(self.hfd)
		sptext:send('txtx', self.hfd)
		sptext:send('txtsz', {self.hfd:getDimensions()})
	end

	love.graphics.setCanvas(canvt)

	-- love.graphics.setFont(fontsm)
	-- love.graphics.print(love.timer.getFPS())
	-- love.graphics.print(('%0.2f %0.2f'):format(scale.x, scale.y))

	if not self.start then
		love.graphics.setFont(font)
		love.graphics.setColor({1, 1, 1})
		love.graphics.printf('click me', 0, FSIZE, IWIDTH, 'center')
	else
		local dy = IHEIGHT + math.max(self.text_timer, FSIZE * -1.5)
		sptext:send('pos', dy)
		love.graphics.setFont(font)
		love.graphics.setColor({1, 1, 0.5})
		love.graphics.setShader(sptext)
		for _ = 1, 100 do
			love.graphics.ellipse('fill', math.random(IWIDTH), math.random(FSIZE * 1.5) + dy, 2, 2)
		end
		love.graphics.setShader()
	end
	love.graphics.setCanvas()
end


function card:explode(roc, expl)
	local tim = twod(50) + 50
	local color = {}

	for i = 1, 3 do
		color[i] = math.random(100) / 100
	end
	color[math.random(3)] = 1

	for _ = 1, twod(30) + 40 do
		if #self.circles < 2000 then
			local dir = math.random() * 2 * math.pi
			local mag = twod(70)
			local cir = table.remove(tq) or {}
			cir.x = roc.x
			cir.y = roc.y
			cir.dx = math.cos(dir) * mag
			cir.dy = math.sin(dir) * mag - 35
			cir.color = color
			cir.timer = twod(50) + 50

			if expl and math.random(15) == 1 then
				cir.fuse = math.floor(cir.timer * 0.9)
			end

			self.circles[#self.circles+1] = cir
		end
	end

	love.audio.play(self.bangs[math.random(#self.bangs)])
end


local drat = 1
function card:fade_canvas()
	if drat % 2 == 1 then
		canv, canv2 = canv2, canv
		love.graphics.setCanvas(canv)
		love.graphics.clear()
		love.graphics.setShader(darken)
		love.graphics.draw(canv2)
		love.graphics.setShader()
		love.graphics.setCanvas()
	elseif drat == 2 then
		canvt, canvt2 = canvt2, canvt
		love.graphics.setCanvas(canvt)
		love.graphics.clear()
		love.graphics.setShader(darken)
		love.graphics.draw(canvt2)
		love.graphics.setShader()
		love.graphics.setCanvas()
	end

	drat = (drat % 6) + 1
end


function card:launch(x, y)
	self.rockets[#self.rockets+1] = {
		x = IWIDTH / 2, y = IHEIGHT + 100,
		tx = x, ty = y,
		dx = 0, dy = -400,
		delay = 100,
	}
end


function card:load()
	self.bangs = {}
	self.circles = {}
	self.rockets = {}
	self.text_timer = 100 -- 1000

	for i = 1, 3 do
		local fn = ('art/sound/fw_%02d.ogg'):format(i)
		self.bangs[#self.bangs+1] = love.audio.newSource(fn, 'static')
	end
end


function card:mask_text()
	-- Mask the text area.
	if self.hfd then
		local dy = IHEIGHT + math.max(self.text_timer, FSIZE * -1.5)
		love.graphics.setColor({0, 0, 0})
		love.graphics.draw(self.hfd, 2, dy + 2)
		love.graphics.draw(self.hfd, -2, dy - 2)
		love.graphics.setColor({1, 1, 1})
	end
end


local elapsed = 0
function card:update(dt)
	elapsed = elapsed + dt

	self:do_circles(dt)
	self:do_rockets(dt)

	if self.start then
		self.text_timer = self.text_timer - 1
	end
end
